<?php

namespace Drupal\activitypub\Services;

use ActivityPhp\Server;
use Drupal\activitypub\Entity\ActivityPubActivityInterface;
use Drupal\activitypub\Entity\ActivityPubActorInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\user\UserInterface;
use HttpSignatures\Context;
use Nyholm\Psr7\Factory\Psr17Factory;
use phpseclib\Crypt\RSA;
use Psr\Log\LoggerInterface;
use Symfony\Bridge\PsrHttpMessage\Factory\PsrHttpFactory;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;

class ActivityPubUtility implements ActivityPubUtilityInterface {

  /**
   * The image style storage.
   *
   * @var \Drupal\image\ImageStyleStorageInterface
   */
  protected $imageStyleStorage;

  /**
   * The actor storage.
   *
   * @var \Drupal\activitypub\Entity\Storage\ActivityPubActorStorageInterface
   */
  protected $actorStorage;

  /**
   * The activitypub storage.
   *
   * @var \Drupal\activitypub\Entity\Storage\ActivityPubActivityStorageInterface
   */
  protected $activityStorage;

  /**
   * The request stack
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack $requestStack
   */
  protected $requestStack;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * ActivityPubUtility constructor
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The request stack.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, RequestStack $request_stack, ConfigFactoryInterface $config_factory, FileSystemInterface $file_system, LoggerInterface $logger) {
    $this->imageStyleStorage = $entity_type_manager->getStorage('image_style');
    $this->actorStorage = $entity_type_manager->getStorage('activitypub_actor');
    $this->activityStorage = $entity_type_manager->getStorage('activitypub_activity');
    $this->requestStack = $request_stack;
    $this->configFactory = $config_factory;
    $this->fileSystem = $file_system;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public function getActivityPubName(ActivityPubActorInterface $actor) {
    return $actor->getName() . '@' . $this->requestStack->getCurrentRequest()->getHost();
  }

  /**
   * {@inheritdoc}
   */
  public function getActivityPubID(ActivityPubActorInterface $actor) {
    return $this->requestStack->getCurrentRequest()->getSchemeAndHttpHost() . $this->requestStack->getCurrentRequest()->getBasePath() . '/user/' . $actor->getOwnerId() . '/activitypub/' . $actor->getName();
  }

  /**
   * {@inheritdoc}
   */
  public function getActivityPubUserAvatar(UserInterface $user) {
    $config = $this->configFactory->get('activitypub.settings');
    $avatar_user_field = $config->get('avatar_user_field');
    /** @var \Drupal\file\FileInterface $image */
    if (!empty($avatar_user_field) && $user->hasField($avatar_user_field) && ($image = $user->get($avatar_user_field)->entity)) {

      /** @var \Drupal\image\ImageStyleInterface $image_style */
      $image_style = $this->imageStyleStorage->load($config->get('avatar_user_style'));
      $image_url = $image_style->buildUrl($image->getFileUri());

    }
    else {
      $avatar_path = activitypub_default_avatar_path();
      if ($avatar_path == 'assets/avatar.png') {
        $avatar_path = drupal_get_path('module', 'activitypub') . '/' . $avatar_path;
      }
      $image_url = file_create_url($avatar_path);
    }

    return $image_url;
  }

  /**
   * {@inheritdoc}
   */
  public function getServer(array $config = []) {
    $config += [
      'cache' => [
        'ttl' => activitypub_cache_ttl(),
        'stream' => activitypub_cache_path()
      ],
      'logger'   => [
        'driver' => '\Psr\Log\NullLogger'
      ],
    ];

    // Override for tests.
    if ($test_prefix = drupal_valid_test_ua()) {
      $config['cache'] = ['enabled' => FALSE];
      $config['http']['agent'] = drupal_generate_test_ua($test_prefix);
    }

    $config['instance']['types'] = 'ignore';
    return new Server($config);
  }

  /**
   * {@inheritdoc}
   */
  public function alterNodeInfo(array &$data) {
    $data['protocols'] = ['activitypub'];
    $data['usage']['users']['total'] = $this->actorStorage->getActorCount(['status' => 1]);
    $data['usage']['localPosts'] = $this->activityStorage->getActivityCount(['collection' => ['outbox', 'liked'], 'status' => 1]);
  }

  /**
   * {@inheritdoc}
   */
  public function onEntityDelete(EntityInterface $entity) {
    if ($entity instanceof ContentEntityInterface && !($entity instanceof ActivityPubActivityInterface)) {
      if ($entity instanceof UserInterface) {
        $actors = $this->actorStorage->loadByProperties(['uid' => $entity->id()]);
        $this->actorStorage->delete($actors);
      }
      elseif ($entity instanceof ActivityPubActorInterface) {
        $this->deleteKeys($entity->getName());
        $activities = $this->activityStorage->loadByProperties(['uid' => $entity->getOwnerId()]);
        $this->activityStorage->delete($activities);
      }
      elseif (method_exists($entity, 'getOwnerId')) {
        // Find a single activity for this entity.
        $conditions = [
          'uid' => $entity->getOwnerId() ,
          'entity_id' => $entity->id(),
          'entity_type_id' => $entity->getEntityTypeId(),
          'collection' => ActivityPubActivityInterface::OUTBOX
        ];
        $activities = $this->activityStorage->loadByProperties($conditions);
        if (count($activities) == 1) {
          /** @var \Drupal\activitypub\Entity\ActivityPubActivityInterface $activity */
          $activity = array_shift($activities);
          $activity->onEntityDelete($entity);
          $activity->delete();
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getOutboxIgnoreTypes() {
    return ['Follow', 'Undo', 'Delete'];
  }

  /**
   * {@inheritdoc}
   */
  public function getTimelineTypes() {
    return ['Create', 'Like', 'Announce'];
  }

  /**
   * {@inheritdoc}
   */
  public function getPublicKey($path) {
    $key_path = rtrim(activitypub_keys_path(), '/') . '/' . $path . '/public.pem';
    return file_get_contents($key_path);
  }

  /**
   * {@inheritdoc}
   */
  public function getPrivateKey($path) {
    $key_path = rtrim(activitypub_keys_path(), '/') . '/' . $path . '/private.pem';
    return file_get_contents($key_path);
  }

  /**
   * {@inheritdoc}
   */
  public function createSignature($private_key_path, $host, $path, $digest, $date = NULL) {
    if (!isset($date)) {
      $date = gmdate('D, d M Y H:i:s T', time());
    }

    $rsa = new RSA();
    $rsa->loadKey($this->getPrivateKey($private_key_path));
    $plaintext = "(request-target): post $path\nhost: $host\ndate: $date\ndigest: $digest";
    $rsa->setHash("sha256");
    $rsa->setSignatureMode(RSA::SIGNATURE_PKCS1);
    return $rsa->sign($plaintext);
  }

  /**
   * {@inheritdoc}
   */
  public function createDigest($message) {
    return 'SHA-256=' . base64_encode(hash('sha256', $message, true));
  }

  /**
   * {@inheritdoc}
   */
  public function generateKeys($path) {
    $return = FALSE;

    try {

      $rsa = new RSA();
      $keys = $rsa->createKey(4096);

      $dir_path = rtrim(activitypub_keys_path(), '/') . '/' . $path;
      $this->fileSystem->prepareDirectory($dir_path, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);

      foreach (['public' => 'publickey', 'private' => 'privatekey'] as $filename => $rsakey) {

        // Key uri.
        $key_uri = "$dir_path/$filename.pem";

        // Write key content to key file.
        $this->fileSystem->saveData($keys[$rsakey], $key_uri, FileSystemInterface::EXISTS_REPLACE);

        // Set correct permission to key file.
        $this->fileSystem->chmod($key_uri, 0600);

      }

      $return = TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error('Error generating keys: @message', ['@message' => $e->getMessage()]);
    }

    return $return;

  }

  /**
   * {@inheritdoc}
   */
  public function deleteKeys($path) {
    $dir_path = rtrim(activitypub_keys_path(), '/') . '/' . $path;
    $this->fileSystem->deleteRecursive($dir_path);
  }

  /**
   * {@inheritdoc}
   */
  public function verifySignature(Request $request, string $actor) {
    $verified = FALSE;

    try {
      $actor = $this->getServer()->actor($actor);
      $keyId = $actor->get('publicKey')['id'];
      $publicKeyPem = $actor->getPublicKeyPem();
      $context = new Context([
        'keys' => [
          $keyId => $publicKeyPem,
        ]
      ]);

      $psr17Factory = new Psr17Factory();
      $psrHttpFactory = new PsrHttpFactory($psr17Factory, $psr17Factory, $psr17Factory, $psr17Factory);
      $psrRequest = $psrHttpFactory->createRequest($request);
      if ($context->verifier()->isSignedWithDigest($psrRequest)) {
        $verified = TRUE;
      }
      else {
        $verified = $context->verifier()->isSigned($psrRequest);
      }
    }
    catch (\Exception $e) {
      $store = TRUE;
      $message = $e->getMessage();
      foreach ($this->ignoreSignatureExceptionMessages() as $m) {
        if (strpos($message, $m) !== FALSE) {
          $store = FALSE;
          break;
        }
      }

      if ($store) {
        $this->logger->error('Signature verifying exception: @message', ['@message' => $message]);
      }
    }

    return $verified;
  }

  /**
   * Returns exceptions we'll ignore.
   *
   * @return string[]
   */
  protected function ignoreSignatureExceptionMessages() {
    return [
      // Client error: `GET example.com` resulted in a `410 Gone` response: {"error":"Gone"}
      '410 Gone',
    ];
  }

}
